/* -*- Mode: java; tab-width: 8; c-basic-offset: 4 -*-
 * Copyright  1997, 1998 Netscape Communications Corporation, All Rights Reserved.
 */

package com.netscape.javascript.CORBA;

import com.netscape.javascript.*;
import org.omg.CORBA.TCKind;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.Any;
import org.omg.CORBA.portable.*;
import org.omg.CORBA.CompletionStatus;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class reflects a IDL structures into the JS
 * environment.
 *
 * @author Ruslan Belkin
 */

public class   Struct  extends Simple   {

    public Struct (Any any)
    {
    	super (any);
    }
    
    public Struct (TypeCode tc)
    {
    	super (tc);
    }
    
    public String getClassName ()   
    {
        return any.type ().kind ().value () == TCKind._tk_except ? "Exception" : "Struct";
    }
   
    public String toString ()
    {
        String s = getClassName () + ":";

        try {
            s += any.type ().id ();
        }
        catch (Exception e) {   s+= "unknown";  };

        s += "={";       
        for (Enumeration e = hashtbl.keys () ; ; )
        {
            Object o = e.nextElement ();
            s += o + "=" + hashtbl.get (o);
            if (!e.hasMoreElements ())
                break;
            s += ";";
        }

        s += "}";
        return s;
    }

    public Object valueOf ()
    {
        return this;
    }

    public boolean has (int index, Scriptable start)
    { 
        try {
            return index >= 0 && index < any.type ().member_count ();
        }
        catch (org.omg.CORBA.UserException ue)  { /* should normally never happen */    }
        return false;
    }

    public boolean has (String id, Scriptable start)
    { 
        return id != null && hashtbl.containsKey (id);
    }

    public void put (int index, Scriptable start, Object value)
    {
        if (!has (index, start))
            return;

        TypeCode tc = any.type ();
        
        if (index < members.length)
        {
            Object current = members[index];
            members[index] = value;
            try {
                encode ();
            }
            catch (org.omg.CORBA.SystemException se)
            {
                members[index] = current;
                Init.raise (new SystemException (se));
            }
        }
    }

    public void put (String id, Scriptable start, Object value)
    {
        if (!has (id, start))
            return;

        TypeCode tc = any.type ();
        
        try {
            int index = 0;
            for ( ; index < members.length && !id.equals (tc.member_name (index)); index++);

            if (index < members.length)
            {
                Object current = members[index];
                members[index] = value;
                try {
                    encode ();
                    hashtbl.put (tc.member_name (index), members[index]);
                }
                catch (org.omg.CORBA.SystemException se)
                {
                    members[index] = current;
                    Init.raise (new SystemException (se));
                }
            }
        }
        catch (org.omg.CORBA.UserException e)  { /* should never happen */ }
    }

    public Object  get (int index, Scriptable start)
    {
        return has (index, start) ?  members[index] : Scriptable.NOT_FOUND;
    }
    
    public Object  get (String id, Scriptable start)
    {
        return has (id, start) ? hashtbl.get (id) : Scriptable.NOT_FOUND;
    }
    
    public  static Object extractObject (InputStream is, TypeCode tc)
        throws org.omg.CORBA.MARSHAL
    {
        switch (tc.kind().value ())
        {
            case TCKind._tk_except :
            case TCKind._tk_struct :

                Any any = Init.orb.create_any ();
                any.type (tc);

                OutputStream os = any.create_output_stream ();
                copyObject (is, os, tc);
                any.read_value (os.create_input_stream (), tc);
                return new Struct (any);
            default :
                return Marshal.extractObject (is, tc);
        }
    }

    public static void copyObject (InputStream is, OutputStream os, TypeCode tc)
        throws org.omg.CORBA.MARSHAL
    {
        int tkind = tc.kind().value();

        try {
            switch (tkind)
            {
                case TCKind._tk_except :
                case TCKind._tk_struct :
                    // copy repository ID
                    if (tkind == TCKind._tk_except)
                        os.write_string (is != null ? is.read_string() : tc.id ());
                    
                    for (int i = 0; i < tc.member_count (); i++)
                        copyObject (is, os, tc.member_type (i));
                    break;
                    
                default:
                    Marshal.copyObject (is, os, tc);
            }
        }
        catch (org.omg.CORBA.UserException e)  { /* should never happen */ }
    }

    public static void writeObject (OutputStream os, TypeCode tc, Object value)
        throws org.omg.CORBA.MARSHAL
    {
        int tkind  = tc.kind().value ();
        boolean ok = true;

        try {
            switch (tkind)
            {
                case TCKind._tk_except :
                case TCKind._tk_struct :
                    if (ok = value instanceof Struct)
                    {
                        if (tkind == TCKind._tk_except)
                            os.write_string (tc.id ());

                        for (int i = 0; i < tc.member_count (); i++)
                            writeObject (os, tc.member_type (i), ((Struct)value).get (i, null));
                    }
                    break;
                default: Marshal.writeObject (os, tc, value);
            }
        }
        catch (org.omg.CORBA.UserException e)   { /* should never happend */ }

        if (!ok)
            throw new org.omg.CORBA.MARSHAL ("cannot cast value: " + value + " to type: " + tc, 0, CompletionStatus.COMPLETED_NO);
    }

    public void decode ()
        throws org.omg.CORBA.MARSHAL
    {
        TypeCode tc = any.type ();

        try {
            InputStream is = any.create_input_stream ();

            members = new Object [tc.member_count ()];
            hashtbl = new Hashtable (tc.member_count ());
                    
            for (int i = 0; i < members.length; i++)
            {
                 members[i] = extractObject (is, tc.member_type (i));
                 hashtbl.put (tc.member_name (i), members[i]);
            }
        }
        catch (org.omg.CORBA.UserException e)   { /* normally should never happend */  }
    }
   
    public static boolean handledType (int tkind)
    {
        switch (tkind)
        {
            case TCKind._tk_except :
            case TCKind._tk_struct :
                return true;
        }
        return false;
    }
    
    Object [] members;
    Hashtable hashtbl;
